
from .tools import get_authors_files_paths, get_target_files_paths
from .vectorizer import vectorize_files
from .analysers import analyse_vectorized_files

from sklearn.neighbors import KNeighborsClassifier
from sklearn.metrics import accuracy_score


class Costyl:

    def __init__(self):
        pass

    def import_authors_files(self, source_folder_path: str):
        """
            This function gets the source folder path and returns the paths of the files and their respective labels(author names)
        """

        self.__source_files_paths, self.__source_files_labels = get_authors_files_paths(source_folder_path)

    def import_target_files(self, source_folder_path: str):
        """
            This function gets the target folder path and returns the paths of the files and their respective labels(file names)
        """

        self.__target_files_paths, self.__target_files_labels = get_target_files_paths(source_folder_path)

    def __vectorize_files(self, source_files_paths: list[str], target_files_paths: list[str]):
        """
            This function vectorizes the source files and target files
        """
        self.__vectorized_source_files_whole, self.__vectorized_source_files_lines = vectorizer.vectorize_files(source_files_paths)
        self.__vectorized_target_files_whole, self.__vectorized_target_files_lines = vectorizer.vectorize_files(target_files_paths)

    def __generate_features(self):
        self.__source_files_data_matrix = analyse_vectorized_files(self.__vectorized_source_files_whole, self.__vectorized_source_files_lines)
        self.__target_files_data_matrix = analyse_vectorized_files(self.__vectorized_target_files_whole, self.__vectorized_target_files_lines)

    def generate_model(self):
        # TODO: Implement the model generation
        self.__vectorize_files(self.__source_files_paths, self.__target_files_paths)
        self.__generate_features()

        self.__model = KNeighborsClassifier(metric="cityblock", n_neighbors=1, algorithm="brute")
        self.__model.fit(self.__source_files_data_matrix, self.__source_files_labels)

    def predict_authors(self):

        predicted_authors = []

        print(self.__target_files_data_matrix)
        input()

        predicted_authors = self.__model.predict(self.__target_files_data_matrix)

        self.__labels_predicted_authors = dict(zip(self.__target_files_labels, predicted_authors))

        return self.__labels_predicted_authors




import math
import numpy as np

import re

from sklearn.feature_extraction.text import CountVectorizer


CPP_KEYWORDS = [
    "alignas",
    "alignof",
    "and",
    "and_eq",
    "asm",
    "atomic_cancel",
    "atomic_commit",
    "atomic_noexcept",
    "auto",
    "bitand",
    "bitor",
    "bool",
    "break",
    "case",
    "catch",
    "char",
    "char8_t",
    "char16_t",
    "char32_t",
    "class",
    "compl",
    "concept",
    "const",
    "consteval",
    "constexpr",
    "constinit",
    "const_cast",
    "continue",
    "co_await",
    "co_return",
    "co_yield",
    "decltype",
    "default",
    "delete",
    "do",
    "double",
    "dynamic_cast",
    "else",
    "enum",
    "explicit",
    "export",
    "extern",
    "false",
    "float",
    "for",
    "friend",
    "goto",
    "if",
    "inline",
    "int",
    "long",
    "mutable",
    "namespace",
    "new",
    "noexcept",
    "not",
    "not_eq",
    "nullptr",
    "operator",
    "or",
    "or_eq",
    "private",
    "protected",
    "public",
    "reflexpr",
    "register",
    "reinterpret_cast",
    "requires",
    "return",
    "short",
    "signed",
    "sizeof",
    "static",
    "static_assert",
    "static_cast",
    "struct",
    "switch",
    "synchronized",
    "template",
    "this",
    "thread_local",
    "throw",
    "true",
    "try",
    "typedef",
    "typeid",
    "typename",
    "union",
    "unsigned",
    "using",
    "virtual",
    "void",
    "volatile",
    "wchar_t",
    "while",
    "xor",
    "xor_eq",
]

CPP_OPERATORS = [
    # Arithmetic operators
    "=",
    "+",
    "++",
    "+=",
    "-",
    "--",
    "-=",
    "*",
    "*=",
    "/",
    "/=",
    "%",
    "%=",
    # Logical operators
    "&&",
    "||",
    "!",
    # Relational operators
    "==",
    "!=",
    ">",
    "<",
    ">=",
    "<=",
    # Bitwise operators
    "&",
    "|",
    "^",
    "~",
    "<<",
    ">>",
    "&=",
    "|=",
    "^=",
    "<<=",
    ">>=",
]

# CPP_ALL_TOKENS = CPP_KEYWORDS + CPP_OPERATORS

def cleanLabelsCountDict(labelsCountDict):

    # This function has the sole purpose of removing the double counts of operators
    # Example: "=" is counted thrice, once for "=" and twice more for "=="
    # "=" is counted twice in another cases, once for "=" and once for "+="

    for oldKey in labelsCountDict.keys():
        for newKey, count in labelsCountDict.items():
            if oldKey in newKey and oldKey != newKey:
                labelsCountDict[oldKey] -= (count * newKey.count(oldKey))

    return labelsCountDict



def lexical_analysis(files: list[str]) -> np.ndarray:

    feats = np.zeros((len(files), 3))

    keywordVectorizer = CountVectorizer(vocabulary=CPP_KEYWORDS)
    operatorVectorizer = CountVectorizer(ngram_range=(1, 4), analyzer='char', vocabulary=CPP_OPERATORS)

    keywordLabels = keywordVectorizer.get_feature_names_out()
    keywordCount = keywordVectorizer.fit_transform(files)
    keywordCount = keywordCount.toarray()

    operatorLabels = operatorVectorizer.get_feature_names_out()
    operatorCount = operatorVectorizer.fit_transform(files)
    operatorCount = operatorCount.toarray()


    for i in range(len(files)):

        # this is needed before counting to clean the keyword double counts
        keywordLabelsCountDict = dict(zip(keywordLabels, keywordCount[i]))
        keywordLabelsCountDict = cleanLabelsCountDict(keywordLabelsCountDict)
        keywordCountSum = sum(list(keywordLabelsCountDict.values()))

        # this is needed before counting to clean the operator double counts
        operatorLabelsCountDict = dict(zip(operatorLabels, operatorCount[i]))
        operatorLabelsCountDict = cleanLabelsCountDict(operatorLabelsCountDict)
        operatorCountSum = sum(list(operatorLabelsCountDict.values()))

        # Not at its most efficient state. Detects strings in commented lines
        stringLiteralCountSum = len(re.findall('(?:"(?:\\\\.|[^"\\\\])*"|\'(?:\\\\.|[^\'\\\\])*\')', files[i]))

        feats[i, 0] = math.log(keywordCountSum/len(files[i]))
        feats[i, 1] = math.log(operatorCountSum/len(files[i]))
        feats[i, 2] = math.log(stringLiteralCountSum/len(files[i]))

    return feats




import glob
import os

def get_authors_files_paths(path: str):
    if os.name == "posix":
        files_paths = glob.glob(path.replace("'", "") + "/*/*.cpp")
        files_labels = [path.split("/")[-2] for path in files_paths]
        return files_paths, files_labels

    elif os.name == "nt":
        files_paths = glob.glob(path.replace("'", "") + "\*\*.cpp")
        files_labels = [path.split("\\")[-2] for path in files_paths]
        return files_paths, files_labels

def get_target_files_paths(path: str):
    if os.name == "posix":
        files_paths = glob.glob(path.replace("'", "") + "/*.cpp")
        files_labels = [path.split("/")[-1].removesuffix(".cpp") for path in files_paths]
        return files_paths, files_labels

    elif os.name == "nt":
        files_paths = glob.glob(path.replace("'", "") + "\*.cpp")
        files_labels = [path.split("\\")[-1].removesuffix(".cpp") for path in files_paths]
        return files_paths, files_labels


def printTestsPredicts(y_test, predictedAuthor):
    print('\n\n' + "\u0332".join("Tests:") + '\n')
    print("Author        || Predicted    : Check status")
    print("--------------------------------------------")
    erros = 0
    for a, b in zip(y_test, predictedAuthor):

        if a == b:
            print(f'{a}{" " * (13 - len(a))} == {b}{" " * (13 - len(b))}: V')
        if a != b:
            print(f'{a}{" " * (13 - len(a))} != {b}{" " * (13 - len(b))}: F')
            erros += 1

    print(f'\nTotal: {len(y_test)}')
    print(f'Erros: {erros}')

def determine_algorithm():
    #TODO implement this function
    pass



from .layout import layout_analysis
from .lexical import lexical_analysis

import numpy as np


def analyse_vectorized_files(vectorized_files_whole: list[str], vectorized_files_lines: list[list[str]]) -> np.ndarray:
    """
    Analyse the files and return a matrix with the results of the analysis
    """
    layout = layout_analysis(vectorized_files_lines)
    lexical = lexical_analysis(vectorized_files_whole)

    return np.concatenate((layout, lexical), axis=1)




from . import __vectorizer as vec

def vectorize_files(source_files_paths: list[str], type: str = "all"):

    if type == "all":
        vectorized_files_whole = [vec.vectorize_file_whole(path) for path in source_files_paths]
        vectorized_files_lines = [vec.vectorize_file_in_lines(path) for path in source_files_paths]

        return vectorized_files_whole, vectorized_files_lines

    elif type == "whole":
        vectorized_files_whole = [vec.vectorize_file_whole(path) for path in source_files_paths]

        return vectorized_files_whole

    elif type == "lines":
        vectorized_source_files_lines = [vec.vectorize_file_in_lines(path) for path in source_files_paths]

        return vectorized_source_files_lines




import io
from sklearn.feature_extraction.text import CountVectorizer


def vectorize_file_whole(filePath: str) -> str:

    with open(filePath, "rb") as f:
        file = f.read().decode("utf-8")

        return file


def vectorize_file_in_lines(filePath: str) -> list[str]:

    with open(filePath, "rb") as f:
        file = f.readlines()
        file = [line.decode("utf-8") for line in file]

        return file




import math
import numpy as np


def layout_analysis(files: list[list[str]]) -> np.ndarray:

    """A function that does a layout analysis of a set of files

        Info inside the returned matrix:

        0 - A int boolean(0(false), 1(true)) representing the use of tabs or spaces for identation. 1 is for the use of tabs, 0 is
        for the use of spaces.

        1 - Log of the number of spaces divided by the file length in characters. Ln(num/len)

        2 - Log of the number of empty lines divided by the file length in characters. Ln(num/len)

        3 - Log of the number of comments divided by the file length in characters. Ln(num/len)

        4 - Average line length

        Returns:
            NDArray[float64]: A matrix containing the info about the analyzed files
    """

    feats = np.zeros((len(files),5))


    for index, file in enumerate(files):
        tabIdent = 0
        spaceCount = 0
        emptyLinesCount = 0
        commentCount = 0
        linesLengthSum = 0

        for line in file:

            linesLengthSum += len(line)

            if "\t" in line:
                tabIdent = 1

            if line == "\n":
                emptyLinesCount += 1

            if "//" in line:
                if '"' in line[0:line.rindex("//")] and '"' in line[line.rindex("//"):]:
                    pass
                else:
                    commentCount += 1

            if "/*" in line:
                commentCount +=1

            for c in line:
                if c == " ":
                    spaceCount += 1



        feats[index, 0] = tabIdent

        if spaceCount != 0:
            feats[index, 1] = math.log(spaceCount/len(file))

        if emptyLinesCount != 0:
            feats[index, 2] = math.log(emptyLinesCount/len(file))

        if commentCount != 0:
            feats[index, 3] = math.log(commentCount/len(file))

        if linesLengthSum != 0:
            feats[index, 4] = linesLengthSum/len(file)

        index += 1

    return feats

if __name__ == "__main__":
    from vectorizer import vectorizeFileInLines
    vecFile = vectorizeFileInLines("/Users/Jbernardis/Documents/stylometryResearch/data/dataset-v3/Benq/Benq-a1-2022-r3.cpp")
    print(layout_analysis([vecFile]))